/*
 * @(#)java_md.c	1.9 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <jni.h>
#include "java.h"

#include "service.h"

#ifdef DEBUG
#define JVM_DLL "jvm_g.dll"
#define JAVA_DLL "java_g.dll"
#else
#define JVM_DLL "jvm.dll"
#define JAVA_DLL "java.dll"
#endif

//01.03.2001 Erwin Vervondel
//Allow definition of jrepath in ini file
extern LPTSTR global_jrePath;
extern LPTSTR global_jvmtype;

/*
 * Prototypes.
 */
static jboolean
GetPublicJREHome(char *path, jint pathsize);

/*
 * Load JVM of "jvmtype", and intialize the invocation functions.  Notice that
 * if jvmtype is NULL, we try to load hotspot, client or classic VM as the default.  Maybe we
 * need an environment variable that dictates the choice of default VM.
 * As of version 0.99l jvmtype can be a comma seperated list. Effectively making
 * the above hotspot default obsolete as they can now be specified explicitely in order of perference
 */
jboolean
LoadJavaVM(char *jvmtype, InvocationFunctions *ifn)
{
    char home[MAXPATHLEN], javadll[MAXPATHLEN], jvmdll[MAXPATHLEN];
    HINSTANCE handle;
    struct stat s;
	size_t pathSize = 32767;
	size_t i;
	LPTSTR path;
	LPTSTR* jvmtypes;

	//01.03.2001 Erwin Vervondel
	//Allow definition of jrepath in ini file

	/* Is JRE defined in ini file? */
	if ( bDebug )
		logToFile( "Searching for JVM path in ini file" );
    if ( strlen( global_jrePath ) > 0 )
	{
		(void) strcpy(home, global_jrePath);
	    sprintf(javadll, "%s\\bin\\" JAVA_DLL, home);
	    if (stat(javadll, &s) == 0)
	       goto jrefound;
    }

	if ( strlen( global_jrePath ) > 0  )
	{
		(void) strcpy(home, global_jrePath);
	    sprintf(javadll, "%s\\jre\\bin\\" JAVA_DLL, home);
	    if (stat(javadll, &s) == 0)
		{
			strcat(home, "\\jre");
			if ( bDebug )
				logToFile( "Found JVM path in ini file" );
			goto jrefound;
		}
    }
	//end 01.03.2001 Erwin Vervondel

    /* Is JRE co-located with the application? */
	if ( bDebug )
		logToFile( "Searching for co-located JVM" );
    if (GetApplicationHome(home, sizeof(home))) {
		sprintf(javadll, "%s\\bin\\" JAVA_DLL, home);
		if (stat(javadll, &s) == 0) 
		{
			if ( bDebug )
				logToFile( "Found co-located JVM" );
			goto jrefound;
		}
	}
    /* Does this app ship a private JRE in <apphome>\jre directory? */
	if ( bDebug )
		logToFile( "Searching JVM in sub dirrecory" );
    sprintf(javadll, "%s\\jre\\bin\\" JAVA_DLL, home);
    if (stat(javadll, &s) == 0) 
	{
		strcat(home, "\\jre");
		if ( bDebug )
			logToFile( "Found JVM in sub dirrecory" );
		goto jrefound;
    }
    /* Look for a public JRE on this machine. */
	if ( bDebug )
		logToFile( "Searching JVM in registry" );
    if (!GetPublicJREHome(home, sizeof(home))) 
	{
		return JNI_FALSE;
    }
	if ( bDebug )
		logToFile( "Found JVM in registry" );

    /* Now we know where JRE is -- the value in the "home" variable. */
jrefound:    

	//2009.10.24 Fix for msvcrt71.dll dependency. Add JVM home to PATH
	//Should be limited to 32bit
	
	path = malloc( pathSize );
	#pragma warning(disable:4267)
	GetEnvironmentVariable( "Path", path, pathSize );
	#pragma warning(default:4267)
	//printf("Path: %s\n", path);
	strcat_s( path, pathSize, ";" );
	strcat_s( path, pathSize, home );
	strcat_s( path, pathSize, "\\bin" );
	SetEnvironmentVariable( "Path", path );

	printf("Path for JVM: %s\n", path);

	//Allow jvmtype to be specified from ini file
	if ( strlen( global_jvmtype ) > 0 )
	{
		jvmtype = global_jvmtype; 
	}
	
	//First we run through the list of jvmtypes until we find one that is installed	
	jvmtypes = stringToArray( jvmtype, ',' );
	if ( jvmtypes != NULL )
	{
		jvmtype = NULL;
		for( i=0; jvmtype==NULL && jvmtypes[i]!=NULL; i++)
		{
			sprintf(jvmdll, "%s\\bin\\%s\\" JVM_DLL, home, jvmtypes[i]);
			if (stat(jvmdll, &s) >= 0) 
			{
				jvmtype = jvmtypes[i];
				if ( bDebug )
				{
					logToFile( "Found jvmtype:" );
					logToFile( jvmdll );
				}
			}
		}
		if ( jvmtype == NULL )
		{
			fprintf(stderr, "Error trying to locate any of the JVM types listed below\n");
			for( i=0; jvmtypes[i]!=NULL; i++)
				fprintf(stderr, "Failed to find jvmtype = %s\n", jvmtypes[i] );
			jvmtype = "notinstalledJVMType";
		}
	}
	//Now we check for some defaults. Should be obsolete if users use the jvmtype list correctly
	//We don't fall back to default if the jvmtype was explictly specified
	else
	{
		sprintf(jvmdll, "%s\\bin\\hotspot\\" JVM_DLL, home);
		if (stat(jvmdll, &s) >= 0) 
		{
			jvmtype = "hotspot";
		} else {
			sprintf(jvmdll, "%s\\bin\\client\\" JVM_DLL, home);
			if (stat(jvmdll, &s) >= 0) 
			{
				jvmtype = "client";
			} else {
				jvmtype = "classic";
			}
		}
	}
		
	/* Now we know what the jvmtype should be */
    sprintf(jvmdll, "%s\\bin\\%s\\" JVM_DLL, home, jvmtype);
    if (bDebug) {
		fprintf(stderr, "Path to JVM DLL is %s\n", jvmdll);
    }
	

    /* Load the Java VM DLL */
    if ((handle = LoadLibrary(jvmdll)) == 0) {
	fprintf(stderr, "Error loading: %s\n", jvmdll);
	return JNI_FALSE;
    }

    /* Now get the function addresses */
    ifn->CreateJavaVM =
	(void *)GetProcAddress(handle, "JNI_CreateJavaVM");
    ifn->GetDefaultJavaVMInitArgs =
	(void *)GetProcAddress(handle, "JNI_GetDefaultJavaVMInitArgs");
    if (ifn->CreateJavaVM == 0 ||
	ifn->GetDefaultJavaVMInitArgs == 0) {
	fprintf(stderr, "Can't find JNI interfaces in: %s\n", jvmdll);
	return JNI_FALSE;
    }

    return JNI_TRUE;
}

/*
 * Get the path to the file that has the usage message for -X options.
 */
void
GetXUsagePath(char *buf, jint bufsize)
{
    GetModuleFileName(GetModuleHandle(JVM_DLL), buf, bufsize);
    *(strrchr(buf, '\\')) = '\0';
    strcat(buf, "\\Xusage.txt");
}

/*
 * If app is "c:\foo\bin\javac", then put "c:\foo" into buf.
 */
jboolean
GetApplicationHome(char *buf, jint bufsize)
{
    char *cp;
    GetModuleFileName(0, buf, bufsize);
    *strrchr(buf, '\\') = '\0'; /* remove .exe file name */
    if ((cp = strrchr(buf, '\\')) == 0) {
	/* This happens if the application is in a drive root, and
	 * there is no bin directory. */
	buf[0] = '\0';
	return JNI_FALSE;
    }
    *cp = '\0';  /* remove the bin\ part */
    return JNI_TRUE;
}

#ifdef JAVAW
__declspec(dllimport) char **__initenv;

int WINAPI
WinMain(HINSTANCE inst, HINSTANCE previnst, LPSTR cmdline, int cmdshow)
{
    __initenv = _environ;
    return main(__argc, __argv);
}
#endif

/*
 * Helpers to look in the registry for a public JRE.
 */
#define DOTRELEASE  "1.2" /* Same for 1.2.1, 1.2.2 etc. */
#define ALTDOTRELEASE  "1.3" 
#define ALT2DOTRELEASE  "1.4" 
#define ALT3DOTRELEASE  "1.5" 
#define ALT4DOTRELEASE  "1.6" 
#define JRE_KEY	    "Software\\JavaSoft\\Java Runtime Environment"

static jboolean
GetStringFromRegistry(HKEY key, const char *name, char *buf, jint bufsize)
{
    DWORD type, size;

    if (RegQueryValueEx(key, name, 0, &type, 0, &size) == 0
	&& type == REG_SZ
	&& (size < (unsigned int)bufsize)) {
	if (RegQueryValueEx(key, name, 0, 0, buf, &size) == 0) {
	    return JNI_TRUE;
	}
    }
    return JNI_FALSE;
}

static jboolean
GetPublicJREHome(char *buf, jint bufsize)
{
    HKEY key, subkey;
    char version[MAXPATHLEN];

    /* Find the current version of the JRE */
    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, JRE_KEY, 0, KEY_READ, &key) != 0) {
	fprintf(stderr, "Error opening registry key '" JRE_KEY "'\n");
	return JNI_FALSE;
    }

    if (!GetStringFromRegistry(key, "CurrentVersion",
			       version, sizeof(version))) {
	fprintf(stderr, "Failed reading value of registry key:\n\t"
		JRE_KEY "\\CurrentVersion\n");
	RegCloseKey(key);
	return JNI_FALSE;
    }

    if ( (strcmp(version, DOTRELEASE) != 0) && 
		 (strcmp(version, ALTDOTRELEASE) != 0) && 
		 (strcmp(version, ALT2DOTRELEASE) != 0) && 
		 (strcmp(version, ALT3DOTRELEASE) != 0) && 
		 (strcmp(version, ALT4DOTRELEASE) != 0) 
	   )
	{
	//Invalid JVM version is no longer a error, just a warning.
	fprintf(stderr, "WARNING: Registry key '" JRE_KEY "\\CurrentVersion'\nhas "
		"value '%s', but only '" DOTRELEASE " or " ALTDOTRELEASE " or " ALT2DOTRELEASE " or " ALT3DOTRELEASE " or " ALT4DOTRELEASE "' has been tested.\n", version);
	//RegCloseKey(key);
	//return JNI_FALSE;
    }

    /* Find directory where the current version is installed. */
    if (RegOpenKeyEx(key, version, 0, KEY_READ, &subkey) != 0) {
	fprintf(stderr, "Error opening registry key '"
		JRE_KEY "\\%s'\n", version);
	RegCloseKey(key);
	return JNI_FALSE;
    }

    if (!GetStringFromRegistry(subkey, "JavaHome", buf, bufsize)) {
	fprintf(stderr, "Failed reading value of registry key:\n\t"
		JRE_KEY "\\%s\\JavaHome\n", version);
	RegCloseKey(key);
	RegCloseKey(subkey);
	return JNI_FALSE;
    }

    if (debug) {
	char micro[MAXPATHLEN];
	if (!GetStringFromRegistry(subkey, "MicroVersion", micro,
				   sizeof(micro))) {
	    printf("Warning: Can't read MicroVersion\n");
	    micro[0] = '\0';
	}
	printf("Version major.minor.micro = %s.%s\n", version, micro);
    }

    RegCloseKey(key);
    RegCloseKey(subkey);
    return JNI_TRUE;
}

/*
 * Support for doing cheap, accurate interval timing.
 */
static jboolean counterAvailable = JNI_FALSE;
static jboolean counterInitialized = JNI_FALSE;
static LARGE_INTEGER counterFrequency;

jlong CounterGet()
{
    LARGE_INTEGER count;

    if (!counterInitialized) {
	counterAvailable = QueryPerformanceFrequency(&counterFrequency);
	counterInitialized = JNI_TRUE;
    }
    if (!counterAvailable) {
	return 0;
    }
    QueryPerformanceCounter(&count);
    return (jlong)(count.QuadPart);
}

jlong Counter2Micros(jlong counts)
{
    if (!counterAvailable || !counterInitialized) {
	return 0;
    }
    return (counts * 1000 * 1000)/counterFrequency.QuadPart;
}
